/* johcgi: CGI gateway for Jabber-Over-HTTP.
   Copyright (C) 2011 Sergey Poznyakoff

   JOH is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   JOH is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with JOH.  If not, see <http://www.gnu.org/licenses/>. */

#include "joh.h"

int param_names[joh_param_max] = {
	joh_param_id,       
	joh_param_key,      
	joh_param_newkey,   
	joh_param_payload,     
	joh_param_server,
	joh_param_port
};

struct joh_param param[joh_param_max];
int fd;

void
cgi_error(int idcode, const char *text, const char *diag)
{
	if (diag)
		error("%s", diag);
	printf("Set-Cookie: ID=%d:0\r\n", idcode);
	printf("Content-Type: text/xml\r\n\r\n");
	if (text)
		printf("%s\r\n", text);
	exit(0);
}

void
store_param(enum joh_param_key key, char *value, size_t len)
{
	if (len == 0)
		return;
	param[key].size = len;
	param[key].ptr = xmalloc(len + 1);
	memcpy(param[key].ptr, value, len);
	param[key].ptr[len] = 0;
}

void
dump_params()
{
	int i;

	for (i = 0; i < joh_param_max; i++)
		if (param[i].size)
			fprintf(stderr, "%u=%s\n",
				param_names[i], param[i].ptr);
}

void
clear_params()
{
	int i;
	for (i = 0; i < joh_param_max; i++) {
		free(param[i].ptr);
		param[i].ptr = NULL;
		param[i].size = 0;
	}
}

/* XEP-0025:

   The client sends requests with bodies in the following format:

          identifier ; key [ ; new_key] , [xml_body]

   If the identifier is zero, key indicates an initial key. In this case,
   new_key should not be specified, and must be ignored.
*/

void
parse_content(char *content_ptr, size_t content_length)
{
	size_t len;
	char *ptr = content_ptr;

	len = strcspn(ptr, ";");
	if (ptr[len] == 0)
		cgi_error(-2, NULL, "bad format");
	store_param(joh_param_id, ptr, len);
	ptr += len + 1;

	len = strcspn(ptr, ",;");
	if (ptr[len] == 0)
		cgi_error(-2, NULL, "bad format");
	store_param(joh_param_key, ptr, len);
	ptr += len;

	if (*ptr++ == ';') {
		len = strcspn(ptr, ",");
		if (ptr[len] == 0)
			cgi_error(-2, NULL, "bad format");
		store_param(joh_param_newkey, ptr, len);
		ptr += len + 1;
	}

	store_param(joh_param_payload, ptr, content_length - (ptr-content_ptr));
}

static void
fallback_jabber_data()
{
	char *s;

	s = getenv("JOH_JABBER_SERVER_URL");
	if (s) {
		char host[NI_MAXHOST];
		char srv[NI_MAXSERV];
		struct joh_sockaddr *jsa = parse_sockaddr(s, NULL);
		switch (jsa->addr->sa_family) {
		case AF_INET:
		case AF_INET6:
			if (getnameinfo(jsa->addr, jsa->addrlen,
					host, sizeof(host), srv, sizeof(srv),
					NI_NUMERICHOST|NI_NUMERICSERV))
				cgi_error(0,
					  "cannot parse address",
					  "cannot parse address");
			store_param(joh_param_server, host, strlen(host));
			store_param(joh_param_port, srv, strlen(srv));
			break;
			
		default:
			cgi_error(0,
				  "unsupported address family",
				  "unsupported address family");
		}

	} else {
		s = getenv("JOH_JABBER_SERVER");
		if (!s)
			s = JABBER_SERVER_NAME;
		store_param(joh_param_server, s, strlen(s));
		s = getenv("JOH_JABBER_PORT");
		if (!s)
			s = JABBER_PORT_STRING;
		store_param(joh_param_port, s, strlen(s));
	}
}

static int
hex_to_num(int c)
{
	static char digits[] = "0123456789";
	static char dohex[] = "abcdef";
	static char uphex[] = "ABCDEF";
	char *p;
	
	if (p = strchr(digits, c))
		return p - digits;
	if (p = strchr(dohex, c))
		return p - dohex + 10;
	if (p = strchr(uphex, c))
		return p - uphex + 10;
	cgi_error(-2, "bad query", "bad query");
	return 0; /* pacify CC */
}

static void
unescape_query(char *q)
{
	char *p;

	p = strchr(q, '%');
	if (!p)
		return;
	for (q = p;;) {
		if (*q == '%') {
			unsigned char c = hex_to_num(q[1])*16 + hex_to_num(q[2]);
			*p++ = c;
			q += 3;
		} else if (!(*p++ = *q++))
			break;
	}
}
		
static void
jabber_data()
{
	int stop = 0;
	char *query;

	query = getenv("QUERY_STRING");
	if (!query) {
		fallback_jabber_data();
		return;
	}

	do {
		char *val;
		size_t len = strcspn(query, ";");

		if (query[len])
			query[len++] = 0;
		else
			stop = 1;

		val = strchr(query, '=');
		if (val) {
			*val++ = 0;
			unescape_query(query);
			unescape_query(val);

			if (strcmp(query, "server") == 0) {
				char *p = strchr(val, ':');
				if (p) {
					*p++ = 0;
					store_param(joh_param_server, val,
						    strlen(val));
					store_param(joh_param_port, p,
						    strlen(p));
				} else
					store_param(joh_param_server, val,
						    strlen(val));
			} else if (strcmp(query, "port") == 0) {
				store_param(joh_param_port, val,
					    strlen(val));
			}
		}
	} while (!stop);
}


void
joh_connect()
{
	int rc;
	char *s;
	struct joh_sockaddr *jsa;
	struct sockaddr *sa;
	socklen_t salen;
	
	s = getenv("JOH_SERVER_URL");
	if (s)
		jsa = parse_sockaddr(s, NULL);
	else {
		static struct joh_sockaddr_hints
			conn_hints = { JOH_HINT_SERVICE,
				       DEFAULT_PORT_STRING };
		jsa = parse_sockaddr("inet://", &conn_hints);
	}
	sa = jsa->addr;
	salen = jsa->addrlen;
	
	fd = socket(sa->sa_family, SOCK_STREAM, 0);
	if (fd == -1) {
		error("socket: %s", strerror(errno));
		cgi_error(0, "connection error", NULL);
	}
	rc = connect(fd, sa, salen);
	if (rc) {
		error("cannot connect to %s: %s",
		      joh_sockaddr_str(jsa), strerror(errno));
		cgi_error(0, "connection error", NULL);
	}
}

void
resolve_server()
{
	int rc;
	struct addrinfo *res;
	char host[NI_MAXHOST];
	char srv[NI_MAXSERV];
	
	if (!(param[joh_param_server].ptr && param[joh_param_port].ptr))
		return;
		
	rc = getaddrinfo(param[joh_param_server].ptr,
			 param[joh_param_port].ptr,
			 NULL, &res);
	if (rc) {
		error("cannot resolve %s:%s: %s",
		      param[joh_param_server].ptr,
		      param[joh_param_port].ptr,
		      gai_strerror(rc));
		cgi_error(-1, NULL, NULL);
	}

	if (getnameinfo(res->ai_addr, res->ai_addrlen,
			host, sizeof(host), srv, sizeof(srv),
			NI_NUMERICHOST|NI_NUMERICSERV))
		cgi_error(-1, NULL, "cannot resolve server address");

	if (strcmp(param[joh_param_server].ptr, host)) {
		free(param[joh_param_server].ptr);
		store_param(joh_param_server, host, strlen(host));
	}
	if (strcmp(param[joh_param_port].ptr, srv)) {
		free(param[joh_param_port].ptr);
		store_param(joh_param_port, srv, strlen(srv));
	}
}

int
main(int argc, char **argv)
{
	size_t content_length;
	char *content_ptr;
	char *s;
	joh_cbuf_t ibuf;
	struct joh_obuf obuf;
	ssize_t rc;
	char buf[IOBUFSIZE];
	
	set_program_name(argv[0]);

	s = getenv("CONTENT_LENGTH");
	if (!s)
		cgi_error(-2, NULL, NULL);

	content_length = strtoul(s, &s, 10);
	if (*s)
		cgi_error(-2, NULL, NULL);
	
	content_ptr = xmalloc(content_length + 1);
	fread(content_ptr, content_length, 1, stdin);
	content_ptr[content_length] = 0;
	parse_content(content_ptr, content_length);
	free(content_ptr);

	if (param[joh_param_id].size == 0)
		cgi_error(-2, NULL, "ID not set");

	jabber_data();
	resolve_server();
	joh_connect();

	/* Send out the pack */
	memset(&obuf, 0, sizeof(obuf));

	joh_pdu_encode(&obuf, joh_param_max, param_names, param);

	while (obuf.off < obuf.size) {
		rc = send(fd, obuf.base + obuf.off,
			  obuf.size - obuf.off,
			  0);
		if (rc < 0) {
			error("I/O error: %s", strerror(errno));
			cgi_error(-1, NULL, NULL);
		}
		obuf.off += rc;
	}
	free(obuf.base);

	/* Read and decode reply */
	ibuf = joh_cbuf_create();
	while ((rc = recv(fd, buf, sizeof(buf), 0)) > 0)
		joh_cbuf_append(ibuf, buf, rc);

	clear_params();
	if (joh_pdu_decode(param, ibuf, "input"))
		cgi_error(0, NULL, "malformed input packet");

	/* Format HTTP reply */
	printf("Set-Cookie: ID=%s\r\n",
	       param[joh_param_id].ptr ? param[joh_param_id].ptr : "0");
	printf("Content-Type: text/xml\r\n\r\n");
	if (param[joh_param_payload].size)
		fwrite(param[joh_param_payload].ptr, 
		       param[joh_param_payload].size,
		       1, stdout);
	return 0;
}
