/* johd: Main Jabber-Over-HTTP daemon.
   Copyright (C) 2011 Sergey Poznyakoff

   JOH is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   JOH is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with JOH.  If not, see <http://www.gnu.org/licenses/>. */

#include "joh.h"
#include <unistd.h>

struct joh_sockaddr *listen_addrs;
struct joh_sockaddr *jabber_addr;

void
version()
{
	printf("%s (%s) %s\n", program_name, PACKAGE, PACKAGE_VERSION);
	printf("Copyright (C) 2011 Sergey Poznyakoff");
	puts("\
\n\
License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n\
This is free software: you are free to change and redistribute it.\n\
There is NO WARRANTY, to the extent permitted by law.\n\
\n\
");
	exit(EX_OK);
}

void
help()
{
	printf("Usage: johd [OPTIONS]\n");
	printf("johd provides a jabber-over-HTTP tunnel (XEP-0025)\n\n");
	printf("OPTIONS:\n");
	puts("\
  -d NUMBER        set debugging level\n\
  -i               show source line information with debug messages\n\
  -l URL           listen on the given URL (several -l options are OK)\n\
  -s URL           jabber server\n\
  -t CLASS:TIMEOUT set session idle timeout\n\
\n\
  -V               print the program version\n\
  -h               show this help summary\n\
\n");
	printf("Report bugs to <%s>.\n", PACKAGE_BUGREPORT);
	exit(EX_OK);
}

static void
set_timeout(char *arg)
{
	enum joh_conntype sesstype;
	time_t timeout;
	int ind = -1;
	static struct { char s[3]; unsigned f; } tab[] = {
		{ "hH", 3600 },
		{ "mM", 60 },
		{ "sS", 1 }
	};
	
	switch (*arg) {
	case 'j':
	case 'J':
		sesstype = conn_jabber;
		break;
	case 'c':
	case 'C':
		sesstype = conn_client;
		break;
	default:
		die(EX_USAGE, "bad type in timeout specification: %s", arg);
	}
	if (*++arg != ':')
		die(EX_USAGE, "timeout specification missing colon: %s", arg);
	timeout = 0;
	for (;;) {
		unsigned long t = strtoul(arg + 1, &arg, 10);
		if (*arg) {
			unsigned x = 0;
			
			for (ind++; ind<sizeof(tab)/sizeof(tab[0]);
			     ind++) {
				if (strchr(tab[ind].s, *arg)) {
					x = tab[ind].f;
					break;
				}
			}
			if (x == 0)
				die(EX_USAGE, "unknown time suffix %c", *arg);
			timeout += t * x;
			if (x > 1)
				continue;
		}
		break;
	}
	if (*arg)
		die(EX_USAGE, "invalid timeout specification: %s", arg);
	joh_session_set_timeout(sesstype, timeout);
}

int
main(int argc, char **argv)
{
	int c;
	char *p;
	static struct joh_sockaddr_hints
		listen_hints = { JOH_HINT_BIND|JOH_HINT_SERVICE,
				 DEFAULT_PORT_STRING },
		jabber_hints = { JOH_HINT_SERVICE, JABBER_PORT_STRING };

	set_program_name(argv[0]);
	while ((c = getopt(argc, argv, "d:il:s:t:hV")) != EOF) {
		switch (c) {
		case 'd':
			debug_level = strtoul(optarg, &p, 10);
			if (*p)
				die(EX_USAGE,
				    "debug level must be a number");
			else if (debug_level < 0 || debug_level > 100)
				die(EX_USAGE,
				    "debug level out of range 0..100");
			break;

		case 'i':
			source_info_option = 1;
			break;
			
		case 'l':
			listen_addrs =
				joh_sockaddr_append(listen_addrs,
						parse_sockaddr(optarg,
							       &listen_hints));
			break;
				
		case 's':
			if (jabber_addr)
				die(EX_USAGE,
				    "only one -s option is allowed");
			jabber_addr = parse_sockaddr(optarg,
						     &jabber_hints);
			if (jabber_addr->next) {
				warn("%s resolves to more than one address, "
				     "using %s",
				     optarg, joh_sockaddr_str(jabber_addr));
				joh_sockaddr_free(jabber_addr->next);
				jabber_addr->next = NULL;
			}
			break;

		case 't':
			set_timeout(optarg);
			break;
				
		case 'h':
			help();
			break;

		case 'V':
			version();
			break;

		default:
			exit(EX_USAGE);
		}
	}

	if (!listen_addrs)
		listen_addrs = parse_sockaddr("inet://", &listen_hints);
	if (!jabber_addr) {
		/* FIXME: Use SRV record, if available. Ditto for -l. */
		jabber_addr = parse_sockaddr(JABBER_SERVER, &jabber_hints);
		if (jabber_addr->next) {
			joh_sockaddr_free(jabber_addr->next);
			jabber_addr->next = NULL;
		}
	}

	joh_conntab_init();
	joh_listen_socket_setup(listen_addrs);
	joh_loop();
	joh_listen_socket_cleanup();
	
	return EX_OK;
}
